#include <stdio.h>
#include <windows.h>
/*
   Use either one of the following two header files.
   Each header file defines a particular kind of buffer data
   structure and it defines the two functions putInBuffer()
   (used by the producer) and TakeOutBuffer() (used by the
   consumer) needed for manipulating its kind of buffer.
*/
#include "StackBuffer.h"
//#include "CircularBuffer.h"

#define BUFFER_COUNT 3
#define P_THREAD_COUNT 3
#define C_THREAD_COUNT 3
#define THREAD_COUNT P_THREAD_COUNT + C_THREAD_COUNT


// Global variables

// You need to declare an array of either mutexes or binary
// semaphores to handle the critical sections for each buffer.

// You need to declare arrays of counting semaphores to bound
// the size of the buffers. Each buffer needs two semaphores,
// one to keep track of open positions in the buffer and another
// to keep track of filled positions in the buffer.

// Array of Buffer data structures.
Buffer buffers[BUFFER_COUNT];

static int runFlag = TRUE;


int main( int argc, char *argv[])
{
   // Producer and consumer thread function prototypes.
   DWORD WINAPI producer(LPVOID);
   DWORD WINAPI consumer(LPVOID);

   HANDLE threadHandle;
   DWORD threadID;
   int producerParam[P_THREAD_COUNT];  // thread function parameters
   int consumerParam[C_THREAD_COUNT];

   int i;

   // Call the function that initializes all of the buffers.
   initializeBuffers(buffers, BUFFER_COUNT);


   // You need to create the mutexes or binary semaphores needed
   // to protect each buffer's critical section.

   // You need to create the two arrays of counting semaphores
   // needed to bound the buffers so that the buffers do not
   // overflow or underflow with data. The global constant SIZE
   // (defined in the header file) specifies how many items each
   // buffer can hold. The buffers are initially empty.


   // Start the consumer threads. Pass the i'th consumer thread the
   // thread parameter 'A'+i (so the consumer threads have thread
   // parameters 'A', 'B', 'C', etc.). Be sure to check that the
   // consumer threads were created without any error. You do not
   // need to save the threadID information.

   // Start the producer threads. Pass the i'th producer thread the
   // thread parameter '1'+i (so the producer threads have thread
   // parameters '1', '2', '3', etc.). Be sure to check that the
   // producer threads were created without any error. You do not
   // need to save the threadID information.


   Sleep(5000);     // let the threads run for a while
   runFlag = FALSE; // now stop the threads
   Sleep(200);      // give them time to stop
   printf("\nThreads stopped.\n");
   return 0;
}//main


// Put the thread parameter in a buffer (the producer).
DWORD WINAPI producer(LPVOID threadParam)
{
   int bufferNo;

   // seed the random number generator
   srand( *(int*)threadParam );

   while(runFlag) // put the thread parameter in a buffer
   {
      // Pick a random buffer number between 0 and BUFFER_COUNT-1.
      bufferNo = rand()%BUFFER_COUNT;


     // Below is the call to putInBuffer() that puts an item (the
     // thread parameter) into the buffer. You need to use a mutex
     // or a binary semaphore (along with WaitForSingleObject) to protect
     // this critical section. You also need to use a pair of counting
     // semaphores (and WaitForSingleObject) to keep the producer from
     // overflowing the buffer.


     // begin critical section
      putInBuffer(*(char*)threadParam, &(buffers[bufferNo]), bufferNo);
     // end of critical section


      //Sleep(30);              // sleep a short while
      Sleep(10*(rand()%10));  // sleep a random amount
   }
   return 0;
}//producer


// Remove an item from a buffer (the consumer).
DWORD WINAPI consumer(LPVOID threadParam)
{
   int bufferNo;
   int result;

   // seed the random number generator
   srand( *(int*)threadParam );

   while(runFlag)  // remove an item from a buffer
   {
      // Pick a random buffer number between 0 and BUFFER_COUNT-1.
      bufferNo = rand()%BUFFER_COUNT;


     // Below is the call to takeFromBuffer() that takes an item out
     // of the buffer. You need to use a mutex or a binary semaphore
     // to protect this critical section. You also need to use a pair
     // of counting semaphores to keep the consumer from underflowing
     // the buffer.


     // begin critical section
      result = takeFromBuffer(&(buffers[bufferNo]), bufferNo);
     // end of critical section


      //Sleep(70);              // sleep a short while
      Sleep(10*(rand()%10));  // sleep a random amount
   }
   return 0;
}//consumer
